package aqua.model.inline

import aqua.model.inline.state.{Arrows, Counter, Exports, Mangler}
import aqua.model.inline.Inline.MergeMode.*
import aqua.model.*
import aqua.model.inline.raw.*
import aqua.raw.ops.*
import aqua.raw.value.*
import aqua.types.{ArrayType, LiteralType, OptionType, StreamType}

import cats.Eval
import cats.syntax.traverse.*
import cats.syntax.monoid.*
import cats.syntax.functor.*
import cats.syntax.flatMap.*
import cats.syntax.apply.*
import cats.instances.list.*
import cats.data.{Chain, State, StateT}
import cats.syntax.applicative.*
import scribe.Logging

object RawValueInliner extends Logging {

  import aqua.model.inline.Inline.*

  private[inline] def unfold[S: Mangler: Exports: Arrows](
    raw: ValueRaw,
    propertiesAllowed: Boolean = true
  ): State[S, (ValueModel, Inline)] = for {
    optimized <- StateT.liftF(Optimization.optimize(raw))
    _ <- StateT.liftF(Eval.later(logger.trace("OPTIMIZIED " + optimized)))
    result <- optimized match {
      case VarRaw(name, t) =>
        for {
          exports <- Exports[S].exports
          model = VarModel(name, t, Chain.empty).resolveWith(exports)
        } yield model -> Inline.empty

      case LiteralRaw(value, t) =>
        State.pure(LiteralModel(value, t) -> Inline.empty)

      case alr: ApplyPropertyRaw =>
        ApplyPropertiesRawInliner(alr, propertiesAllowed)

      case cr: CollectionRaw =>
        CollectionRawInliner(cr, propertiesAllowed)

      case dr: MakeStructRaw =>
        MakeStructRawInliner(dr, propertiesAllowed)

      case ar: AbilityRaw =>
        MakeAbilityRawInliner(ar, propertiesAllowed)

      case auor: ApplyUnaryOpRaw =>
        ApplyUnaryOpRawInliner(auor, propertiesAllowed)

      case abbor: ApplyBinaryOpRaw =>
        ApplyBinaryOpRawInliner(abbor, propertiesAllowed)

      case cr: CallArrowRaw =>
        CallArrowRawInliner(cr, propertiesAllowed)

      case cs: CallServiceRaw =>
        CallServiceRawInliner(cs, propertiesAllowed)

    }
  } yield result

  private[inline] def inlineToTree[S: Mangler: Exports: Arrows](
    inline: Inline
  ): State[S, List[OpModel.Tree]] =
    (inline.mergeMode match {
      case SeqMode => SeqModel.wrap(inline.predo) :: Nil
      case ParMode => inline.predo.toList
    }).pure

  private[inline] def toModel[S: Mangler: Exports: Arrows](
    unfoldF: State[S, (ValueModel, Inline)]
  ): State[S, (ValueModel, Option[OpModel.Tree])] =
    for {
      vmp <- unfoldF
      (vm, map) = vmp
      _ = logger.trace("MOD " + vm)
      dc <- Exports[S].exports
      _ = logger.trace("DEC " + dc)

      ops <- inlineToTree(map)
      _ = logger.trace("desugarized ops: " + ops)
      _ = logger.trace("map was: " + map)
    } yield vm -> parDesugarPrefix(ops.filterNot(_ == EmptyModel.leaf))

  def collectionToModel[S: Mangler: Exports: Arrows](
    value: CollectionRaw,
    assignTo: Option[String]
  ): State[S, (ValueModel, Option[OpModel.Tree])] = {
    logger.trace("RAW COLLECTION " + value)
    toModel(CollectionRawInliner.unfoldCollection(value, assignTo))
  }

  def valueToModel[S: Mangler: Exports: Arrows](
    value: ValueRaw,
    propertiesAllowed: Boolean = true
  ): State[S, (ValueModel, Option[OpModel.Tree])] = for {
    _ <- StateT.liftF(Eval.later(logger.trace("RAW " + value)))
    model <- toModel(unfold(value, propertiesAllowed))
  } yield model

  def valueListToModel[S: Mangler: Exports: Arrows](
    values: List[ValueRaw]
  ): State[S, List[(ValueModel, Option[OpModel.Tree])]] =
    values.traverse(valueToModel(_))

  /**
   * Unfold all arguments and make CallModel
   * @param flatStreamArguments canonicalize and flatten all stream arguments if true
   */
  def callToModel[S: Mangler: Exports: Arrows](
    call: Call,
    flatStreamArguments: Boolean
  ): State[S, (CallModel, Option[OpModel.Tree])] =
    valueListToModel(call.args).flatMap { args =>
      if (flatStreamArguments)
        args.map(arg => TagInliner.flat(arg._1, arg._2, true)).sequence
      else
        State.pure(args)
    }.map { list =>
      (
        CallModel(
          list.map(_._1),
          call.exportTo.map(CallModel.callExport)
        ),
        parDesugarPrefix(list.flatMap(_._2))
      )
    }
}
